#include <fcntl.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <linux/if_tun.h>
#include <linux/netlink.h>
#include <linux/rtnetlink.h>
#include <net/if.h>
#include <sys/ioctl.h>

#include "ifnow_tun.h"

// this is called from the ESP Now receive callback
// it passes the contents received from the wifi interface
int ifnow_tun_send(ifnow_tun_cfg_t *tun_cfg, uint8_t *buf, uint8_t len) {
  printf("tun: sending %d bytes to fd: %d\n", len, tun_cfg->fd_tuntap);

  for(uint16_t i = 0; i < len; i++) {
    printf("%02X ", buf[i]);
    if(i % 16 == 15) printf("\n");
  }
  printf("\n");

  return write(tun_cfg->fd_tuntap, buf, len);
}

void *ifnow_tun_thread_handle(void *arg) {
  ifnow_tun_cfg_t *tun_cfg = arg;

  printf("starting ifnow_tun thread.\n");

  ssize_t receive_count = 0;
  uint8_t receive_buffer[IFNOW_MAX_RECEIVE];

  while(1) {
    receive_count = read(tun_cfg->fd_tuntap, &receive_buffer, IFNOW_MAX_RECEIVE);

    if(receive_count > 0 && tun_cfg->receive != NULL) {
      tun_cfg->receive(tun_cfg, (uint8_t*)&receive_buffer, receive_count);
    }
  };
}

int ifnow_tun_netlink_connect(ifnow_tun_cfg_t *tun_cfg) {
  struct sockaddr_nl netlink_socket;

  tun_cfg->fd_netlink = socket(AF_NETLINK, SOCK_RAW | SOCK_CLOEXEC, NETLINK_ROUTE);
  if(tun_cfg->fd_netlink < 0) {
    perror("Failed to get netlink socket");

    return -4;
  }

  memset(&netlink_socket, 0, sizeof(struct sockaddr_nl));
  netlink_socket.nl_family = AF_NETLINK;
  int err = bind(tun_cfg->fd_netlink, (struct sockaddr*) &netlink_socket, sizeof(struct sockaddr_nl));
  if(err < 0) {
    perror("Failed to bind netlink socket");

    return -5;
  }
}

int ifnow_tun_netlink_set_ip(ifnow_tun_cfg_t *tun_cfg) {
  struct {
    struct nlmsghdr  header;
    struct ifaddrmsg content;
    char             attributes_buf[64];
  } request;

  struct rtattr *request_attr;
  size_t attributes_buf_avail = sizeof request.attributes_buf;

  memset(&request, 0, sizeof request);
  request.header.nlmsg_type = RTM_NEWADDR;
  request.header.nlmsg_len = NLMSG_LENGTH(sizeof request.content);
  request.header.nlmsg_flags = NLM_F_REQUEST | NLM_F_EXCL | NLM_F_CREATE;
  request.content.ifa_index = if_nametoindex(tun_cfg->if_name);
  request.content.ifa_family = AF_INET;
  request.content.ifa_prefixlen = 32;

  /* request.attributes[IFA_LOCAL] = address */
  request_attr = IFA_RTA(&request.content);

  request_attr->rta_type = IFA_LOCAL;
  request_attr->rta_len = RTA_LENGTH(sizeof (struct in_addr));
  request.header.nlmsg_len += request_attr->rta_len;
  inet_pton(AF_INET, tun_cfg->addr_s, RTA_DATA(request_attr));

  /* request.attributes[IFA_ADDRESS] = address */
  request_attr = RTA_NEXT(request_attr, attributes_buf_avail);

  request_attr->rta_type = IFA_ADDRESS;
  request_attr->rta_len = RTA_LENGTH(sizeof (struct in_addr));
  request.header.nlmsg_len += request_attr->rta_len;
  inet_pton(AF_INET, tun_cfg->addr_s, RTA_DATA(request_attr));

  if (send(tun_cfg->fd_netlink, &request, request.header.nlmsg_len, 0) == -1) {
    perror("Failed to connect netlink");

    return -6;
  }
}

int ifnow_tun_netlink_up(ifnow_tun_cfg_t *tun_cfg) {
  struct {
    struct nlmsghdr  header;
    struct ifinfomsg content;
  } request;

  printf("ifnow_tun_netlink_up: %s\n", tun_cfg->if_name);

  memset(&request, 0, sizeof request);
  request.header.nlmsg_len = NLMSG_LENGTH(sizeof request.content);
  request.header.nlmsg_flags = NLM_F_REQUEST;
  request.header.nlmsg_type = RTM_NEWLINK;
  request.content.ifi_index = if_nametoindex(tun_cfg->if_name);
  request.content.ifi_flags = IFF_UP;
  request.content.ifi_change = 1;

  if (send(tun_cfg->fd_netlink, &request, request.header.nlmsg_len, 0) == -1) {
    perror("Failed to send netlink up");

    return -1;
  }
  return 0;
}

int ifnow_tun_tuntap_connect(ifnow_tun_cfg_t *tun_cfg) {
  struct ifreq ifr;

  tun_cfg->fd_tuntap = open("/dev/net/tun", O_RDWR | O_CLOEXEC);
  if(tun_cfg->fd_tuntap < 0) {
    perror("Failed to open /dev/net/tun");

    return -2;
  }

  memset(&ifr, 0, sizeof(struct ifreq));
  ifr.ifr_flags = IFF_TAP | IFF_NO_PI;

  if(tun_cfg->if_name != NULL) {
    memcpy(ifr.ifr_name, tun_cfg->if_name, strlen(tun_cfg->if_name) + 1);
  }

  int err = ioctl(tun_cfg->fd_tuntap, TUNSETIFF, &ifr);
  if(err < 0) {
    perror("Failed to set interface name");
    return -3;
  }

}

int ifnow_tun_init(ifnow_tun_cfg_t *tun_cfg) {
  printf("tun_cfg->addr_s: %s\n", tun_cfg->addr_s);

  int err = inet_pton(AF_INET, tun_cfg->addr_s, &tun_cfg->addr);
  if(err < 0) {
    perror("Failed to parse IP address");
    
    return -1;
  }

  ifnow_tun_tuntap_connect(tun_cfg);
  ifnow_tun_netlink_connect(tun_cfg);
  ifnow_tun_netlink_set_ip(tun_cfg);
  // TODO send netlink route request
  ifnow_tun_netlink_up(tun_cfg);

  pthread_create(&tun_cfg->thread_id, NULL, &ifnow_tun_thread_handle, tun_cfg);

  return 0;
}
